// ---------------------------------------------------------------------------
//  Includes
// ---------------------------------------------------------------------------


#ifndef NO_SAX

#include <util/XMLUniDefs.hpp>
#include <sax/AttributeList.hpp>
#include "SAXElemHandler.h"
#include "Globals.h"

// ---------------------------------------------------------------------------
//  Local const data
//
//  Note: This is the 'safe' way to do these strings. If you compiler supports
//        L"" style strings, and portability is not a concern, you can use
//        those types constants directly.
// ---------------------------------------------------------------------------
static const XMLCh  gEndElement[] = { chOpenAngle, chForwardSlash, chNull };
static const XMLCh  gEndPI[] = { chQuestion, chCloseAngle, chNull };
static const XMLCh  gStartPI[] = { chOpenAngle, chQuestion, chNull };
static const XMLCh  gXMLDecl1[] =
{
  chOpenAngle, chQuestion, chLatin_x, chLatin_m, chLatin_l
  ,   chSpace, chLatin_v, chLatin_e, chLatin_r, chLatin_s, chLatin_i
  ,   chLatin_o, chLatin_n, chEqual, chDoubleQuote, chDigit_1, chPeriod
  ,   chDigit_0, chDoubleQuote, chSpace, chLatin_e, chLatin_n, chLatin_c
  ,   chLatin_o, chLatin_d, chLatin_i, chLatin_n, chLatin_g, chEqual
  ,   chDoubleQuote, chNull
};

static const XMLCh  gXMLDecl2[] =
{
  chDoubleQuote, chQuestion, chCloseAngle
  ,   chLF, chNull
};




// ---------------------------------------------------------------------------
//  SAXElemHandlers: Constructors and Destructor
// ---------------------------------------------------------------------------
SAXElemHandlers::SAXElemHandlers( const   char* const              encodingName
				  , const XMLFormatter::UnRepFlags unRepFlags) :

  fFormatter
(
 encodingName
 , this
 , XMLFormatter::NoEscapes
 , unRepFlags
 )
{
  //
  //  Go ahead and output an XML Decl with our known encoding. This
  //  is not the best answer, but its the best we can do until we
  //  have SAX2 support.
  //
  fFormatter << gXMLDecl1 << fFormatter.getEncodingName() << gXMLDecl2;
}

SAXElemHandlers::~SAXElemHandlers()
{
}


// ---------------------------------------------------------------------------
//  SAXElemHandlers: Overrides of the output formatter target interface
// ---------------------------------------------------------------------------
void SAXElemHandlers::writeChars(const XMLByte* const toWrite)
{
}

void SAXElemHandlers::writeChars(const XMLByte* const toWrite,
				 const unsigned int count,
				 XMLFormatter* const formatter)
{
  // For this one, just dump them to the standard output
  // Surprisingly, Solaris was the only platform on which
  // required the char* cast to print out the string correctly.
  // Without the cast, it was printing the pointer value in hex.
  // Quite annoying, considering every other platform printed
  // the string with the explicit cast to char* below.
  //  cout.write((char *) toWrite, (int) count);
  //	cout.flush();
}


// ---------------------------------------------------------------------------
//  SAXElemHandlers: Overrides of the SAX ErrorHandler interface
// ---------------------------------------------------------------------------
void SAXElemHandlers::error(const SAXParseException& e)
{
  //    cerr << "\nError at file " << StrX(e.getSystemId())
  //		 << ", line " << e.getLineNumber()
  //		 << ", char " << e.getColumnNumber()
  //         << "\n  Message: " << StrX(e.getMessage()) << endl;
}

void SAXElemHandlers::fatalError(const SAXParseException& e)
{
  //    cerr << "\nFatal Error at file " << StrX(e.getSystemId())
  //		 << ", line " << e.getLineNumber()
  //		 << ", char " << e.getColumnNumber()
  //        << "\n  Message: " << StrX(e.getMessage()) << endl;
}

void SAXElemHandlers::warning(const SAXParseException& e)
{
  //    cerr << "\nWarning at file " << StrX(e.getSystemId())
  //		 << ", line " << e.getLineNumber()
  //		 << ", char " << e.getColumnNumber()
  //         << "\n  Message: " << StrX(e.getMessage()) << endl;
}


// ---------------------------------------------------------------------------
//  SAXElemHandlers: Overrides of the SAX DTDHandler interface
// ---------------------------------------------------------------------------
void SAXElemHandlers::unparsedEntityDecl(const     XMLCh* const name
					 , const   XMLCh* const publicId
					 , const   XMLCh* const systemId
					 , const   XMLCh* const notationName)
{
  // Not used at this time
}


void SAXElemHandlers::notationDecl(const   XMLCh* const name
				   , const XMLCh* const publicId
				   , const XMLCh* const systemId)
{
  // Not used at this time
}


// ---------------------------------------------------------------------------
//  SAXElemHandlers: Overrides of the SAX DocumentHandler interface
// ---------------------------------------------------------------------------
void SAXElemHandlers::characters(const     XMLCh* const    chars
				 , const   unsigned int    length)
{
  XmlNode *node;
  node = curnode;
  if( node )
    {
      GString str(StrX(chars), length);
      node->content += str;
    }
  /*
    fFormatter.formatBuf(chars, length, XMLFormatter::CharEscapes);
    */
}


void SAXElemHandlers::endDocument()
{
}


void SAXElemHandlers::endElement(const XMLCh* const name)
{

  XmlNode* node = curnode;
  if( node )
    {
      node = node->get_parent();
      curnode = node;
    }
  // No escapes are legal here
  /*
    fFormatter << XMLFormatter::NoEscapes << gEndElement << name << chCloseAngle;
    */
}


void SAXElemHandlers::ignorableWhitespace( const   XMLCh* const chars
					   ,const  unsigned int length)
{
  fFormatter.formatBuf(chars, length, XMLFormatter::NoEscapes);
}


void SAXElemHandlers::processingInstruction(const  XMLCh* const target
                                            , const XMLCh* const data)
{
  fFormatter << XMLFormatter::NoEscapes << gStartPI  << target;
  if (data)
    fFormatter << chSpace << data;
  fFormatter << XMLFormatter::NoEscapes << gEndPI;
}


void SAXElemHandlers::startDocument()
{
}


void SAXElemHandlers::startElement(const   XMLCh* const    name
				   ,       AttributeList&  attributes)
{
  // The name has to be representable without any escapes

  GP<XmlNode> node = new XmlNode(curnode);

  if( !curnode )
    {
      curnode = node;
      root = node;
    }
  else
    {
      ///add node as child
      GP<XmlNode> parent = curnode;
      parent->add_child(node);
    }
  StrX x(name);
  node->name = GString(x).upcase();
  unsigned int len = attributes.getLength();
  for (unsigned int index = 0; index < len; index++)
    {
      StrX attr(attributes.getName(index));
      StrX val(attributes.getValue(index));
      node->attrvalues[GString(attr).upcase()] = val;
    }

  curnode = node;

  /*
    fFormatter  << XMLFormatter::NoEscapes
    << chOpenAngle << name;

    unsigned int len = attributes.getLength();
    for (unsigned int index = 0; index < len; index++)
    {
    //
    //  Again the name has to be completely representable. But the
    //  attribute can have refs and requires the attribute style
    //  escaping.
    //
    fFormatter  << XMLFormatter::NoEscapes
    << chSpace << attributes.getName(index)
    << chEqual << chDoubleQuote
    << XMLFormatter::AttrEscapes
    << attributes.getValue(index)
    << XMLFormatter::NoEscapes
    << chDoubleQuote;
    }
    fFormatter << chCloseAngle;
    */
}

#endif
